from enum import Enum
from typing import Optional
from datetime import datetime
from dataclasses import dataclass

import ciso8601
from mashumaro import DataClassJSONMixin
from mashumaro.types import SerializableType


FigiName = str
TickerName = str


class BaseModel(DataClassJSONMixin):
    pass


@dataclass
class Error(BaseModel):
    message: Optional[str] = None
    code: Optional[str] = None


class Status(Enum):
    OK = 'Ok'
    ERROR = 'Error'


class Currency(Enum):
    RUB = 'RUB'
    USD = 'USD'
    EUS = 'EUR'
    GBP = 'GBP'
    HKD = 'HKD'
    CHF = 'CHF'
    JPY = 'JPY'
    CNY = 'CNY'
    TRY = 'TRY'


class InstrumentType(Enum):
    STOCK = 'Stock'
    CURRENCY = 'Currency'
    BOND = 'Bond'
    ETF = 'Etf'


@dataclass
class MoneyAmount(BaseModel):
    currency: Currency
    value: float


class ISODateTime(datetime, SerializableType):

    def _serialize(self):
        return self.isoformat()

    @classmethod
    def _deserialize(cls, value):
        dt = ciso8601.parse_datetime(value)
        return cls(
            dt.year, dt.month, dt.day, dt.hour, dt.minute,
            dt.second, dt.microsecond, dt.tzinfo
        )



from typing import List
from dataclasses import dataclass
from enum import Enum

from tinkoff.investments.model.base import BaseModel


BrokerAccountID = str


class BrokerAccountType(Enum):
    TINKOFF = 'Tinkoff'
    TINKOFF_IIS = 'TinkoffIis'


@dataclass
class UserAccount(BaseModel):
    brokerAccountType: BrokerAccountType
    brokerAccountId: BrokerAccountID


@dataclass
class UserAccounts(BaseModel):
    accounts: List[UserAccount]


__all__ = [
    'BrokerAccountID',
    'BrokerAccountType',
    'UserAccount',
    'UserAccounts',
]



from .rest import TinkoffInvestmentsRESTClient
from .streaming import TinkoffInvestmentsStreamingClient, EventsBroker


__all__ = [
    'TinkoffInvestmentsRESTClient',
    'TinkoffInvestmentsStreamingClient',
    'EventsBroker',
]



from dataclasses import dataclass
from typing import List, Optional
from enum import Enum

from tinkoff.investments.model.base import BaseModel, FigiName


class TradingStatus(Enum):
    NORMAL_TRADING = 'NormalTrading'
    NOT_AVAILABLE_FOR_TRADING = 'NotAvailableForTrading'


@dataclass
class OrderBookEntity(BaseModel):
    price: float
    quantity: int


@dataclass
class OrderBook(BaseModel):
    figi: FigiName
    depth: int
    bids: List[OrderBookEntity]
    asks: List[OrderBookEntity]
    tradeStatus: TradingStatus
    minPriceIncrement: float
    faceValue: Optional[float] = None
    lastPrice: Optional[float] = None
    closePrice: Optional[float] = None
    limitUp: Optional[float] = None
    limitDown: Optional[float] = None


__all__ = [
    'TradingStatus',
    'OrderBookEntity',
    'OrderBook',
]



from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Optional, ClassVar, Type, TypeVar, Any

from mashumaro import DataClassJSONMixin

from tinkoff.base import classproperty
from tinkoff.investments.model.base import BaseModel, FigiName, ISODateTime
from tinkoff.investments.model.market.candles import Candle, CandleResolution
from tinkoff.investments.model.market.orderbook import OrderBook


class EventName(Enum):
    CANDLE = 'candle'
    ORDERBOOK = 'orderbook'
    INSTRUMENT_INFO = 'instrument_info'
    ERROR = 'error'


class BaseEventKey(DataClassJSONMixin):
    event_name: ClassVar[EventName]

    def subscribe_key(self) -> Dict[str, Any]:
        return {'event': f'{self.event_name.value}:subscribe',
                **self.to_dict()}

    def unsubscribe_key(self) -> Dict[str, Any]:
        return {'event': f'{self.event_name.value}:unsubscribe',
                **self.to_dict()}


@dataclass(unsafe_hash=True)
class CandleEventKey(BaseEventKey):
    event_name = EventName.CANDLE
    figi: FigiName
    interval: CandleResolution


@dataclass(unsafe_hash=True)
class OrderBookEventKey(BaseEventKey):
    event_name = EventName.ORDERBOOK
    figi: FigiName
    depth: int


@dataclass(unsafe_hash=True)
class InstrumentInfoKey(BaseEventKey):
    event_name = EventName.INSTRUMENT_INFO
    figi: FigiName


ConcreteEventKey = TypeVar('ConcreteEventKey', bound=BaseEventKey)


class BaseEvent(BaseModel):
    event_name: ClassVar[EventName]

    @classproperty
    def key_type(self) -> Type[ConcreteEventKey]:
        return EventKeyMapping[self.event_name]

    def key(self):
        raise NotImplementedError


@dataclass
class CandleEvent(BaseEvent):
    event_name = EventName.CANDLE
    figi: FigiName
    time: ISODateTime
    interval: CandleResolution
    o: float
    c: float
    h: float
    l: float
    v: int

    def key(self):
        return self.key_type(figi=self.figi, interval=self.interval)

    @classmethod
    def from_candle(cls, candle: Candle) -> 'CandleEvent':
        return CandleEvent(
            figi=candle.figi,
            time=candle.time,
            interval=candle.interval,
            o=candle.o,
            c=candle.c,
            h=candle.h,
            l=candle.l,
            v=candle.v
        )


@dataclass
class OrderBookEvent(BaseEvent):
    event_name = EventName.ORDERBOOK
    figi: FigiName
    depth: int
    # TODO: сделать списком сущностей Order
    bids: List[List[float]]
    asks: List[List[float]]

    def key(self):
        return self.key_type(figi=self.figi, depth=self.depth)

    @classmethod
    def from_orderbook(cls, orderbook: OrderBook) -> 'OrderBookEvent':
        return OrderBookEvent(
            figi=orderbook.figi,
            depth=orderbook.depth,
            bids=[[e.price, e.quantity] for e in orderbook.bids],
            asks=[[e.price, e.quantity] for e in orderbook.asks],
        )


@dataclass
class InstrumentInfoEvent(BaseEvent):
    event_name = EventName.INSTRUMENT_INFO
    figi: FigiName
    min_price_increment: float
    lot: float
    trade_status: str
    accrued_interest: Optional[float] = None
    limit_up: Optional[float] = None
    limit_down: Optional[float] = None

    def key(self):
        return self.key_type(figi=self.figi)


@dataclass
class ErrorEvent(BaseEvent):
    event_name = EventName.ERROR
    error: str
    request_id: Optional[str] = None

    def key(self):
        raise NotImplementedError


@dataclass
class StreamingMessage(DataClassJSONMixin):
    event: EventName
    payload: Dict[Any, Any]

    @property
    def parsed_payload(self):
        return EventMapping[self.event].from_dict(self.payload)


EventMapping = {
    EventName.CANDLE: CandleEvent,
    EventName.ORDERBOOK: OrderBookEvent,
    EventName.INSTRUMENT_INFO: InstrumentInfoEvent,
    EventName.ERROR: ErrorEvent,
}


EventKeyMapping = {
    EventName.CANDLE: CandleEventKey,
    EventName.ORDERBOOK: OrderBookEventKey,
    EventName.INSTRUMENT_INFO: InstrumentInfoKey,
}



from dataclasses import dataclass
from typing import Any

from .base import BaseModel, Status


@dataclass
class TinkoffInvestmentsAPIResponse(BaseModel):
    trackingId: str
    status: Status
    payload: Any

    def is_successful(self):
        return self.status is Status.OK



import asyncio
import logging
from typing import Any, Dict, Type, Optional, Callable, List

from aiohttp import ClientWebSocketResponse, WSMsgType, ClientConnectionError

from tinkoff.base import BaseHTTPClient
from tinkoff.investments.client.environments import Environment, EnvironmentURL
from tinkoff.investments.model.streaming import (
    BaseEvent,
    CandleEvent,
    OrderBookEvent,
    InstrumentInfoEvent,
    BaseEventKey,
    StreamingMessage,
    EventName,
)


logger = logging.getLogger('streaming-api')


class BaseEventStream:
    EVENT_TYPE: Type[BaseEvent] = None

    def __init__(self):
        self._subscribers = {}  # type: Dict[BaseEventKey, Callable]
        self._client = None  # type: Optional[TinkoffInvestmentsStreamingClient]

    def __call__(self, *args, **kwargs):
        def decorator(callback):
            self._subscribers[
                self.EVENT_TYPE.key_type(*args, **kwargs)] = callback
            return callback
        return decorator

    async def subscribe(self, callback, *args, **kwargs):
        key = self.EVENT_TYPE.key_type(*args, **kwargs)
        self._subscribers[key] = callback
        await self._client.request(key.subscribe_key())

    async def unsubscribe(self, *args, **kwargs):
        key = self.EVENT_TYPE.key_type(*args, **kwargs)
        self._subscribers.pop(key, None)
        await self._client.request(key.unsubscribe_key())

    async def publish(self, event: BaseEvent):
        callback = self._subscribers.get(event.key())  # TODO: сделать иначе
        if callback:
            await callback(event)


class CandleEventStream(BaseEventStream):
    EVENT_TYPE = CandleEvent


class OrderBookEventStream(BaseEventStream):
    EVENT_TYPE = OrderBookEvent


class InstrumentInfoEventStream(BaseEventStream):
    EVENT_TYPE = InstrumentInfoEvent


class EventsBroker:
    def __init__(self):
        self.candles = CandleEventStream()
        self.orderbooks = OrderBookEventStream()
        self.instrument_info = InstrumentInfoEventStream()

        self._routes = {
            EventName.CANDLE: self.candles,
            EventName.ORDERBOOK: self.orderbooks,
            EventName.INSTRUMENT_INFO: self.instrument_info,
        }

    def add_publisher(self, client: 'TinkoffInvestmentsStreamingClient'):
        self.candles._client = client
        self.orderbooks._client = client
        self.instrument_info._client = client

    async def publish(self, event: BaseEvent):
        await self._routes[event.event_name].publish(event)


class TinkoffInvestmentsStreamingClient(BaseHTTPClient):
    def __init__(
            self,
            token: str,
            events: EventsBroker = None,
            receive_timeout: Optional[float] = 5,
            heartbeat: Optional[float] = 3,
            reconnect_timeout: float = 3,
    ):
        super().__init__(
            base_url=EnvironmentURL[Environment.STREAMING],
            headers={
                'authorization': f'Bearer {token}'
            }
        )
        self.events = events or EventsBroker()
        self.events.add_publisher(self)
        self._ws = None  # type: Optional[ClientWebSocketResponse]
        self._receive_timeout = receive_timeout
        self._heartbeat = heartbeat
        self._reconnect_timeout = reconnect_timeout

    async def request(self, key: Dict[str, Any]):
        if self._ws:
            await self._ws.send_json(key)

    async def run(self):
        while not self.closed:
            if self._session.closed:
                return
            try:
                async with self._session.ws_connect(
                    url=self._base_url,
                    timeout=0.,
                    receive_timeout=self._receive_timeout,
                    heartbeat=self._heartbeat,
                ) as ws:
                    self._ws = ws
                    await self._run(ws)
            except asyncio.TimeoutError:
                await asyncio.sleep(self._reconnect_timeout)
            except ClientConnectionError:
                await asyncio.sleep(self._reconnect_timeout)

    async def _run(self, ws: ClientWebSocketResponse):
        await self._subscribe_to_streams(ws)
        async for msg in ws:
            # noinspection PyUnresolvedReferences
            if msg.type == WSMsgType.TEXT:
                # noinspection PyUnresolvedReferences
                msg = StreamingMessage.from_json(msg.data)
                try:
                    await self.events.publish(msg.parsed_payload)
                except Exception as e:
                    logger.exception(
                        'Unhandled exception in streaming event handler: %s', e
                    )

    async def _subscribe_to_streams(self, ws: ClientWebSocketResponse):
        coros = (ws.send_json(key) for key in self._subscription_keys())
        await asyncio.gather(*coros)

    @property
    def _event_streams(self):
        return (self.events.candles, self.events.orderbooks,
                self.events.instrument_info)

    def _subscription_keys(self) -> List[Dict[str, Any]]:
        keys = []
        for event_stream in self._event_streams:
            # noinspection PyProtectedMember
            keys.extend([key.subscribe_key()
                         for key in event_stream._subscribers.keys()])
        return keys


__all__ = [
    'EventsBroker',
    'TinkoffInvestmentsStreamingClient'
]



from aiohttp import ClientWebSocketResponse
from tinkoff.base import BaseHTTPClient
from tinkoff.investments.model.base import FigiName
from tinkoff.investments.model.streaming import BaseEvent, BaseEventKey, EventName
from tinkoff.investments.model.market.candles import CandleResolution
from typing import Any, Callable, Dict, List, Optional

class CandleEventStream:
    _subscribers: Dict[BaseEventKey, Callable] = ...
    def __call__(self, figi: FigiName, interval: CandleResolution): ...
    async def subscribe(self, callback, figi: FigiName, interval: CandleResolution) -> None: ...
    async def unsubscribe(self, figi: FigiName, interval: CandleResolution) -> None: ...

class OrderBookEventStream:
    _subscribers: Dict[BaseEventKey, Callable] = ...
    def __call__(self, figi: FigiName, depth: int): ...
    async def subscribe(self, callback, figi: FigiName, depth: int) -> None: ...
    async def unsubscribe(self, figi: FigiName, depth: int) -> None: ...

class InstrumentInfoEventStream:
    _subscribers: Dict[BaseEventKey, Callable] = ...
    def __call__(self, figi: FigiName): ...
    async def subscribe(self, callback, figi: FigiName) -> None: ...
    async def unsubscribe(self, figi: FigiName) -> None: ...

class EventsBroker:
    candles: CandleEventStream = CandleEventStream()
    orderbooks: OrderBookEventStream = OrderBookEventStream()
    instrument_info: InstrumentInfoEventStream = InstrumentInfoEventStream()
    _routes: Dict[EventName, Any] = ...
    def __init__(self) -> None: ...
    def add_publisher(self, client: TinkoffInvestmentsStreamingClient) -> None: ...
    async def publish(self, event: BaseEvent) -> None: ...

class TinkoffInvestmentsStreamingClient(BaseHTTPClient):
    events: EventsBroker = EventsBroker()
    _receive_timeout: Optional[float]
    _heartbeat: Optional[float]
    _reconnect_timeout: float
    def __init__(
            self,
            token: str,
            events: EventsBroker = None,
            receive_timeout: Optional[float] = 5,
            heartbeat: Optional[float] = 3,
            reconnect_timeout: float = 3
    ) -> None: ...
    async def run(self) -> None: ...
    async def _run(self, ws: ClientWebSocketResponse) -> None: ...
    async def _subscribe_to_streams(self, ws: ClientWebSocketResponse) -> None: ...
    @property
    def _event_streams(self):
        return (self.events.candles, self.events.orderbooks,
                self.events.instrument_info)
    def _subscription_keys(self) -> List[Dict[str, Any]]: ...



from dataclasses import dataclass
from enum import Enum
from typing import List, Optional

from tinkoff.investments.model.base import (
    BaseModel,
    ISODateTime,
    MoneyAmount,
    Currency,
    FigiName,
    InstrumentType,
)


OperationID = str
TradeID = str


class OperationStatus(Enum):
    DONE = 'Done'
    DECLINE = 'Decline'
    PROGRESS = 'Progress'


class OperationType(Enum):
    BUY = 'Buy'
    SELL = 'Sell'


class OperationTypeWithCommission(Enum):
    BUY = 'Buy'
    BUY_CARD = 'BuyCard'
    SELL = 'Sell'
    BROKER_COMMISSION = 'BrokerCommission'
    EXCHANGE_COMMISSION = 'ExchangeCommission'
    SERVICE_COMMISSION = 'ServiceCommission'
    MARGIN_COMMISSION = 'MarginCommission'
    OTHER_COMMISSION = 'OtherCommission'
    PAY_IN = 'PayIn'
    PAY_OUT = 'PayOut'
    TAX = 'Tax'
    TAX_LUCRE = 'TaxLucre'
    TAX_DIVIDEND = 'TaxDividend'
    TAX_COUPON = 'TaxCoupon'
    TAX_BACK = 'TaxBack'
    REPAYMENT = 'Repayment'
    PART_REPAYMENT = 'PartRepayment'
    COUPON = 'Coupon'
    DIVIDEND = 'Dividend'
    SECURITY_IN = 'SecurityIn'
    SECURITY_OUT = 'SecurityOut'


@dataclass
class OperationTrade(BaseModel):
    tradeId: TradeID
    date: ISODateTime
    price: float
    quantity: int


@dataclass
class Operation(BaseModel):
    id: OperationID
    status: OperationStatus
    currency: Currency
    payment: float
    isMarginCall: bool
    date: ISODateTime
    trades: Optional[List[OperationTrade]] = None
    commission: Optional[MoneyAmount] = None
    price: Optional[float] = None
    quantity: Optional[int] = None
    figi: Optional[FigiName] = None
    instrumentType: Optional[InstrumentType] = None
    operationType: Optional[OperationTypeWithCommission] = None


@dataclass
class Operations(BaseModel):
    operations: List[Operation]


__all__ = [
    'OperationID',
    'TradeID',
    'OperationStatus',
    'OperationType',
    'OperationTypeWithCommission',
    'OperationTrade',
    'Operation',
    'Operations',
]



from dataclasses import dataclass

from tinkoff.investments.model.base import BaseModel, Currency, FigiName
from tinkoff.investments.model.user.accounts import (
    BrokerAccountID,
    BrokerAccountType,
)


@dataclass
class SandboxAccount(BaseModel):
    brokerAccountType: BrokerAccountType
    brokerAccountId: BrokerAccountID


@dataclass
class SandboxAccountRegisterRequest(BaseModel):
    brokerAccountType: BrokerAccountType


@dataclass
class SandboxSetCurrencyBalanceRequest(BaseModel):
    currency: Currency
    balance: float


@dataclass
class SandboxSetPositionBalanceRequest(BaseModel):
    figi: FigiName
    balance: float


__all__ = [
    'SandboxAccount',
    'SandboxAccountRegisterRequest',
    'SandboxSetCurrencyBalanceRequest',
    'SandboxSetPositionBalanceRequest',
]



from datetime import datetime, timezone
from typing import List

from tinkoff.investments.api.base import BaseTinkoffInvestmentsAPI
from tinkoff.investments.model.base import FigiName
from tinkoff.investments.model.user.accounts import BrokerAccountID
from tinkoff.investments.model.operations import Operation, Operations


class OperationsAPI(BaseTinkoffInvestmentsAPI):
    async def get(
            self,
            dt_from: datetime,
            dt_to: datetime,
            figi: FigiName = None,
            broker_account_id: BrokerAccountID = None
    ) -> List[Operation]:

        if not dt_from.tzinfo:
            dt_from = dt_from.replace(tzinfo=timezone.utc)
        if not dt_to.tzinfo:
            dt_to = dt_to.replace(tzinfo=timezone.utc)
        params = {
            'from': dt_from.isoformat(),
            'to': dt_to.isoformat(),
        }
        if figi is not None:
            params['figi'] = figi
        if broker_account_id is not None:
            params['brokerAccountId'] = broker_account_id

        payload = await self._request(
            method='GET',
            path='/operations',
            params=params,
        )
        return Operations.from_dict(payload).operations



from asyncio import TimeoutError

from tinkoff.investments.model.base import Error, Status


class TinkoffInvestmentsError(Exception):
    pass


class TinkoffInvestmentsUsageError(TinkoffInvestmentsError):
    pass


class TinkoffInvestmentsUnauthorizedError(TinkoffInvestmentsError):
    def __str__(self):
        return 'Have you missed the real token?'


class TinkoffInvestmentsTooManyRequestsError(TinkoffInvestmentsError):
    pass


class TinkoffInvestmentsTimeoutError(TinkoffInvestmentsError, TimeoutError):
    pass


class TinkoffInvestmentsAPIError(TinkoffInvestmentsError):
    def __init__(self, tracking_id: str, status: Status, error: Error):
        self.trackingId = tracking_id
        self.status = status
        self.error = error

    def __str__(self):
        return f'{self.error.code}, {self.error.message}'


__all__ = [
    'TinkoffInvestmentsError',
    'TinkoffInvestmentsUsageError',
    'TinkoffInvestmentsAPIError',
    'TinkoffInvestmentsUnauthorizedError',
    'TinkoffInvestmentsTooManyRequestsError',
    'TinkoffInvestmentsTimeoutError',
]



from .sandbox import SandboxAPI
from .orders import OrdersAPI
from .portfolio import PortfolioAPI
from .market import MarketAPI
from .operations import OperationsAPI
from .user import UserAPI


__all__ = [
    'SandboxAPI',
    'OrdersAPI',
    'PortfolioAPI',
    'MarketAPI',
    'OperationsAPI',
    'UserAPI',
]



import asyncio
from typing import Any, Dict, Optional

from tinkoff.base import BaseHTTPClient, RateLimiter, RateLimitReached
from tinkoff.investments.api import (
    SandboxAPI,
    OrdersAPI,
    PortfolioAPI,
    MarketAPI,
    OperationsAPI,
    UserAPI,
)
from tinkoff.investments.client.environments import Environment, EnvironmentURL
from tinkoff.investments.client.exceptions import (
    TinkoffInvestmentsUnauthorizedError,
    TinkoffInvestmentsTooManyRequestsError,
    TinkoffInvestmentsTimeoutError,
)


class TinkoffInvestmentsRESTClient(BaseHTTPClient):
    def __init__(
            self,
            token: str,
            environment: Environment = Environment.PRODUCTION,
            timeout: Optional[float] = 5,
            wait_on_rate_limit: bool = True,
    ):

        super(TinkoffInvestmentsRESTClient, self).__init__(
            base_url=EnvironmentURL[environment],
            headers={
                'authorization': f'Bearer {token}'
            },
            timeout=timeout,
        )
        self.sandbox = SandboxAPI(self)
        self.orders = OrdersAPI(self)
        self.portfolio = PortfolioAPI(self)
        self.market = MarketAPI(self)
        self.operations = OperationsAPI(self)
        self.user = UserAPI(self)
        self.wait_on_rate_limit = wait_on_rate_limit

    async def _request(self, method, path, rate_limit=None, **kwargs):
        # type: (str, str, Optional[RateLimiter], Any) -> Dict[Any, Any]
        try:
            if rate_limit:
                await rate_limit.acquire(self.wait_on_rate_limit)
            response = await self._session.request(
                method=method,
                url=self._base_url / path.lstrip('/'),
                **kwargs
            )
            if response.status == 401:
                raise TinkoffInvestmentsUnauthorizedError
            elif response.status == 429:
                raise TinkoffInvestmentsTooManyRequestsError
            else:
                # TODO: ловить другие исключения, если в ответе не json
                return await response.json()
        except asyncio.TimeoutError:
            raise TinkoffInvestmentsTimeoutError from None
        except RateLimitReached:
            raise TinkoffInvestmentsTooManyRequestsError from None

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


__all__ = [
    'TinkoffInvestmentsRESTClient'
]



from dataclasses import dataclass
from typing import List
from enum import Enum

from tinkoff.investments.model.base import BaseModel, ISODateTime, FigiName


class CandleResolution(Enum):
    MIN_1 = '1min'
    MIN_2 = '2min'
    MIN_3 = '3min'
    MIN_5 = '5min'
    MIN_10 = '10min'
    MIN_15 = '15min'
    MIN_30 = '30min'
    HOUR = 'hour'
    DAY = 'day'
    WEEK = 'week'
    MONTH = 'month'


@dataclass
class Candle(BaseModel):
    figi: FigiName
    interval: CandleResolution
    # TODO: сделать алиасы на понятные названия
    o: float
    c: float
    h: float
    l: float
    v: int
    time: ISODateTime


@dataclass
class Candles(BaseModel):
    figi: FigiName
    interval: CandleResolution
    candles: List[Candle]


__all__ = [
    'CandleResolution',
    'Candle',
    'Candles',
]



from typing import Union, Dict, List, Any

from tinkoff.base import RateLimiter
from tinkoff.investments.model.base import Error
from tinkoff.investments.model.response import TinkoffInvestmentsAPIResponse
from tinkoff.investments.client.exceptions import TinkoffInvestmentsAPIError


class BaseTinkoffInvestmentsAPI:
    def __init__(self, client):
        self._client = client
        self._rate_limit = RateLimiter(rate=120, period=60)

    async def _request(self, method, path, **kwargs):
        # type: (str, str, Any) -> Union[Dict[Any, Any], List[Any]]
        # noinspection PyProtectedMember
        data = await self._client._request(
            method, path, self._rate_limit, **kwargs
        )
        response = TinkoffInvestmentsAPIResponse.from_dict(data)
        if response.is_successful():
            return response.payload
        else:
            raise TinkoffInvestmentsAPIError(
                tracking_id=response.trackingId,
                status=response.status,
                error=Error(
                    message=response.payload.get('message'),
                    code=response.payload.get('code'),
                )
            )


__all__ = [
    'BaseTinkoffInvestmentsAPI',
]



from typing import List

from tinkoff.investments.api.base import BaseTinkoffInvestmentsAPI
from tinkoff.investments.model.user.accounts import UserAccount, UserAccounts


class UserAPI(BaseTinkoffInvestmentsAPI):
    async def get_accounts(self) -> List[UserAccount]:
        payload = await self._request(
            method='GET',
            path='/user/accounts',
        )
        return UserAccounts.from_dict(payload).accounts



from dataclasses import dataclass
from enum import Enum
from typing import Optional

from tinkoff.investments.model.base import BaseModel, FigiName, MoneyAmount
from tinkoff.investments.model.operations import OperationType


OrderID = str


class OrderStatus(Enum):
    NEW = 'New'
    PARTIALLY_FILL = 'PartiallyFill'
    FILL = 'Fill'
    CANCELLED = 'Cancelled'
    REPLACED = 'Replaced'
    PENDING_CANCEL = 'PendingCancel'
    REJECTED = 'Rejected'
    PENDING_REPLACE = 'PendingReplace'
    PENDING_NEW = 'PendingNew'


class OrderType(Enum):
    LIMIT = 'Limit'
    MARKET = 'Market'


@dataclass
class Order(BaseModel):
    orderId: OrderID
    figi: FigiName
    operation: OperationType
    status: OrderStatus
    requestedLots: int
    executedLots: int
    type: OrderType
    price: float


@dataclass
class LimitOrderRequest(BaseModel):
    lots: int
    operation: OperationType
    price: float


@dataclass
class PlacedLimitOrder(BaseModel):
    orderId: OrderID
    operation: OperationType
    status: OrderStatus
    requestedLots: int
    executedLots: int
    rejectReason: Optional[str] = None
    message: Optional[str] = None
    commission: Optional[MoneyAmount] = None


@dataclass
class MarketOrderRequest(BaseModel):
    lots: int
    operation: OperationType


@dataclass
class PlacedMarketOrder(BaseModel):
    orderId: OrderID
    operation: OperationType
    status: OrderStatus
    requestedLots: int
    executedLots: int
    rejectReason: Optional[str] = None
    message: Optional[str] = None
    commission: Optional[MoneyAmount] = None


__all__ = [
    'OrderID',
    'OrderStatus',
    'OrderType',
    'Order',
    'LimitOrderRequest',
    'PlacedLimitOrder',
    'MarketOrderRequest',
    'PlacedMarketOrder',
]



from dataclasses import dataclass
from typing import List, Optional

from tinkoff.investments.model.base import (
    BaseModel,
    Currency,
    MoneyAmount,
    FigiName,
    TickerName,
    InstrumentType,
)


@dataclass
class PortfolioPosition(BaseModel):
    figi: FigiName
    instrumentType: InstrumentType
    balance: float
    lots: int
    name: str
    ticker: Optional[TickerName] = None
    isin: Optional[str] = None
    blocked: Optional[float] = None
    expectedYield: Optional[MoneyAmount] = None
    averagePositionPrice: Optional[MoneyAmount] = None
    averagePositionPriceNoNkd: Optional[MoneyAmount] = None


@dataclass
class CurrencyPosition(BaseModel):
    currency: Currency
    balance: float
    blocked: Optional[float] = None


@dataclass
class Portfolio(BaseModel):
    positions: List[PortfolioPosition]


@dataclass
class Currencies(BaseModel):
    currencies: List[CurrencyPosition]


__all__ = [
    'InstrumentType',
    'PortfolioPosition',
    'CurrencyPosition',
    'Portfolio',
    'Currencies',
]



from dataclasses import dataclass
from typing import Optional, List

from tinkoff.investments.model.base import (
    BaseModel,
    Currency,
    FigiName,
    TickerName,
    InstrumentType,
)


@dataclass
class MarketInstrument(BaseModel):
    figi: FigiName
    ticker: TickerName
    lot: int
    name: str
    type: InstrumentType
    currency: Optional[Currency] = None
    isin: Optional[str] = None
    minPriceIncrement: Optional[float] = None


@dataclass
class MarketInstrumentList(BaseModel):
    total: int
    instruments: List[MarketInstrument]


__all__ = [
    'MarketInstrument',
    'MarketInstrumentList',
]



from tinkoff.investments.api.base import BaseTinkoffInvestmentsAPI
from tinkoff.investments.model.base import Currency, FigiName
from tinkoff.investments.model.sandbox import (
    SandboxAccountRegisterRequest,
    SandboxAccount,
    SandboxSetCurrencyBalanceRequest,
    SandboxSetPositionBalanceRequest,
)
from tinkoff.investments.model.user.accounts import (
    BrokerAccountType,
    BrokerAccountID,
)


class SandboxAccountCurrenciesAPI(BaseTinkoffInvestmentsAPI):
    async def set_balance(self, broker_account_id, currency, balance):
        # type: (BrokerAccountID, Currency, float) -> None
        await self._request(
            method='POST',
            path='/sandbox/currencies/balance',
            params={
                'brokerAccountId': broker_account_id
            },
            json=SandboxSetCurrencyBalanceRequest(
                currency=currency,
                balance=balance
            ).to_dict()
        )


class SandboxAccountPositionsAPI(BaseTinkoffInvestmentsAPI):
    async def set_balance(self, broker_account_id, figi, balance):
        # type: (BrokerAccountID, FigiName, float) -> None
        await self._request(
            method='POST',
            path='/sandbox/positions/balance',
            params={
                'brokerAccountId': broker_account_id
            },
            json=SandboxSetPositionBalanceRequest(
                figi=figi,
                balance=balance
            ).to_dict()
        )


class SandboxAccountsAPI(BaseTinkoffInvestmentsAPI):
    def __init__(self, *args, **kwargs):
        super(SandboxAccountsAPI, self).__init__(*args, **kwargs)
        self.currencies = SandboxAccountCurrenciesAPI(*args, **kwargs)
        self.positions = SandboxAccountPositionsAPI(*args, **kwargs)

    async def register(self, broker_account_type=BrokerAccountType.TINKOFF):
        # type: (BrokerAccountType) -> SandboxAccount
        payload = await self._request(
            method='POST',
            path='/sandbox/register',
            json=SandboxAccountRegisterRequest(
                brokerAccountType=broker_account_type
            ).to_dict(),
        )
        return SandboxAccount.from_dict(payload)

    async def remove(self, broker_account_id: BrokerAccountID):
        await self._request(
            method='POST',
            path='/sandbox/remove',
            params={
                'brokerAccountId': broker_account_id
            }
        )

    async def clear(self, broker_account_id: BrokerAccountID):
        await self._request(
            method='POST',
            path='/sandbox/clear',
            params={
                'brokerAccountId': broker_account_id
            }
        )


class SandboxAPI(BaseTinkoffInvestmentsAPI):
    def __init__(self, *args, **kwargs):
        super(SandboxAPI, self).__init__(*args, **kwargs)
        self.accounts = SandboxAccountsAPI(*args, **kwargs)



from .client.environments import Environment
from .client.rest import TinkoffInvestmentsRESTClient
from .client.streaming import TinkoffInvestmentsStreamingClient, EventsBroker
from .model.base import (
    FigiName, TickerName, Currency, InstrumentType, MoneyAmount
)
from .model.market.candles import CandleResolution, Candle
from .model.market.orderbook import TradingStatus, OrderBookEntity, OrderBook
from .model.market.instruments import MarketInstrument
from .model.orders import (
    OrderID, OrderStatus, OrderType, Order, PlacedMarketOrder, PlacedLimitOrder
)
from .model.operations import (
    OperationID, TradeID, OperationStatus, OperationType,
    OperationTypeWithCommission, OperationTrade, Operation
)
from .model.portfolio import PortfolioPosition, CurrencyPosition
from .model.user.accounts import BrokerAccountID, BrokerAccountType, UserAccount
from .model.sandbox import SandboxAccount
from .model.streaming import CandleEvent, OrderBookEvent, InstrumentInfoEvent



from typing import List
from tinkoff.investments.api.base import BaseTinkoffInvestmentsAPI


from tinkoff.investments.model.user.accounts import BrokerAccountID
from tinkoff.investments.model.portfolio import (
    PortfolioPosition,
    CurrencyPosition,
    Portfolio,
    Currencies,
)


class PortfolioAPI(BaseTinkoffInvestmentsAPI):
    async def get_positions(self, broker_account_id=None):
        # type: (BrokerAccountID) -> List[PortfolioPosition]
        if broker_account_id is not None:
            params = {'brokerAccountId': broker_account_id}
        else:
            params = {}
        payload = await self._request(
            method='GET',
            path='/portfolio',
            params=params,
        )
        return Portfolio.from_dict(payload).positions

    async def get_currencies(self, broker_account_id=None):
        # type: (BrokerAccountID) -> List[CurrencyPosition]
        if broker_account_id is not None:
            params = {'brokerAccountId': broker_account_id}
        else:
            params = {}
        payload = await self._request(
            method='GET',
            path='/portfolio/currencies',
            params=params,
        )
        return Currencies.from_dict(payload).currencies



#!/usr/bin/env python

from setuptools import setup, find_packages


setup(
    name="tinkoff-api",
    version="1.0",
    description="Python Tinkoff API client for asyncio and humans",
    long_description=open('README.md', encoding='utf8').read(),
    long_description_content_type='text/markdown',
    platforms="all",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Topic :: Office/Business :: Financial",
        "Topic :: Office/Business :: Financial :: Investment",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 4 - Beta",
    ],
    license="Apache License, Version 2.0",
    author="Alexander Tikhonov",
    author_email="random.gauss@gmail.com",
    url='https://github.com/Fatal1ty/tinkoff-api',
    package_data={
        "tinkoff": [
            "py.typed",
            "investments/client/streaming.pyi"
        ]
    },
    packages=find_packages(exclude=("tests",)),
    python_requires=">=3.6",
    install_requires=[
        "aiohttp>=3.6.2,<3.7",
        "yarl>=1.4.2,<1.5",
        "mashumaro>=1.12,<2.0",
        "ciso8601==2.1.2",
    ]
)



from typing import List

from tinkoff.investments.api.base import BaseTinkoffInvestmentsAPI
from tinkoff.investments.model.user.accounts import BrokerAccountID
from tinkoff.investments.model.orders import (
    Order,
    OrderID,
    PlacedLimitOrder,
    LimitOrderRequest,
    PlacedMarketOrder,
    MarketOrderRequest,
)
from tinkoff.investments.model.operations import OperationType
from tinkoff.investments.model.base import FigiName


class OrdersAPI(BaseTinkoffInvestmentsAPI):
    async def get(self, broker_account_id=None):
        # type: (BrokerAccountID) -> List[Order]

        if broker_account_id is not None:
            params = {'brokerAccountId': broker_account_id}
        else:
            params = {}

        payload = await self._request(
            method='GET',
            path='/orders',
            params=params
        )
        return [Order.from_dict(obj) for obj in payload]

    async def create_limit_order(
            self,
            figi: FigiName,
            lots: int,
            operation: OperationType,
            price: float,
            broker_account_id: BrokerAccountID = None
    ) -> PlacedLimitOrder:

        params = {
            'figi': figi
        }
        if broker_account_id is not None:
            params['brokerAccountId'] = broker_account_id

        payload = await self._request(
            method='POST',
            path='/orders/limit-order',
            params=params,
            json=LimitOrderRequest(
                lots=lots,
                operation=operation,
                price=price,
            ).to_dict()
        )
        return PlacedLimitOrder.from_dict(payload)

    async def create_market_order(
            self,
            figi: FigiName,
            lots: int,
            operation: OperationType,
            broker_account_id: BrokerAccountID = None
    ) -> PlacedMarketOrder:

        params = {
            'figi': figi
        }
        if broker_account_id is not None:
            params['brokerAccountId'] = broker_account_id

        payload = await self._request(
            method='POST',
            path='/orders/market-order',
            params=params,
            json=MarketOrderRequest(
                lots=lots,
                operation=operation,
            ).to_dict()
        )
        return PlacedMarketOrder.from_dict(payload)

    async def cancel(self, order_id, broker_account_id=None):
        # type: (OrderID, BrokerAccountID) -> None

        params = {
            'orderId': order_id
        }
        if broker_account_id is not None:
            params['brokerAccountId'] = broker_account_id

        await self._request(
            method='POST',
            path='/orders/cancel',
            params=params,
        )



import time
import asyncio
from collections import deque
from typing import Optional, Dict, Deque

import aiohttp
from yarl import URL


class BaseHTTPClient:
    def __init__(self, base_url, headers=None, timeout=None):
        # type: (URL, Optional[Dict[str, str]], Optional[float]) -> None
        self._base_url = base_url
        self.__headers = headers
        self.__timeout = timeout
        self.__session = None  # type: Optional[aiohttp.ClientSession]

        self._closed = False

    @property
    def _session(self):
        if not self.__session:
            self.__session = aiohttp.ClientSession(
                headers=self.__headers,
                timeout=aiohttp.ClientTimeout(total=self.__timeout)
            )
        return self.__session

    async def close(self):
        self._closed = True
        await self._session.close()

    @property
    def closed(self):
        return self._closed


# noinspection PyPep8Naming
class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


class RateLimitReached(Exception):
    pass


class RateLimiter:
    def __init__(self, rate: int, period: float):
        self.rate = rate
        self.period = period
        self.request_times: Deque[float] = deque()

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    async def acquire(self, blocking=True):
        while not self._try_to_acquire():
            if not blocking:
                raise RateLimitReached
            await asyncio.sleep(0.25)

    def _try_to_acquire(self):
        now = time.monotonic()
        while self.request_times:
            if now - self.request_times[0] > self.period:
                self.request_times.popleft()
            else:
                break
        if len(self.request_times) < self.rate:
            self.request_times.append(now)
            return True
        else:
            return False


__all__ = [
    'BaseHTTPClient',
    'classproperty',
    'RateLimiter',
    'RateLimitReached',
]



from enum import Enum

from yarl import URL


class Environment(Enum):
    SANDBOX = 'sandbox'
    PRODUCTION = 'production'
    STREAMING = 'streaming'


EnvironmentURL = {
    Environment.SANDBOX: URL('https://api-invest.tinkoff.ru/openapi/sandbox/'),
    Environment.PRODUCTION: URL('https://api-invest.tinkoff.ru/openapi/'),
    Environment.STREAMING: URL(
        'wss://api-invest.tinkoff.ru/openapi/md/v1/md-openapi/ws'),
}


__all__ = [
    'Environment',
    'EnvironmentURL',
]



from datetime import datetime, timezone
from typing import Optional, List, Any

from tinkoff.investments.api.base import BaseTinkoffInvestmentsAPI

from tinkoff.investments.client.exceptions import (
    TinkoffInvestmentsUsageError,
    TinkoffInvestmentsAPIError,
)
from tinkoff.investments.model.base import FigiName, TickerName
from tinkoff.investments.model.market.orderbook import OrderBook
from tinkoff.investments.model.market.candles import (
    Candle,
    Candles,
    CandleResolution,
)
from tinkoff.investments.model.market.instruments import (
    MarketInstrument,
    MarketInstrumentList,
)


class MarketInstrumentsAPI(BaseTinkoffInvestmentsAPI):
    async def search(self, ticker: TickerName) -> List[MarketInstrument]:
        return await self.__get_instruments(
            path='/market/search/by-ticker',
            ticker=ticker,
        )

    async def get(self, figi: FigiName) -> Optional[MarketInstrument]:
        try:
            payload = await self._request(
                method='GET',
                path='/market/search/by-figi',
                params={'figi': figi},
            )
            return MarketInstrument.from_dict(payload)
        except TinkoffInvestmentsAPIError as e:
            if e.error.code == 'NOT_FOUND':
                return None
            else:
                raise e from None

    async def get_stocks(self) -> List[MarketInstrument]:
        return await self.__get_instruments('/market/stocks')

    async def get_bonds(self) -> List[MarketInstrument]:
        return await self.__get_instruments('/market/bonds')

    async def get_etfs(self) -> List[MarketInstrument]:
        return await self.__get_instruments('/market/etfs')

    async def get_currencies(self) -> List[MarketInstrument]:
        return await self.__get_instruments('/market/currencies')

    async def __get_instruments(self, path, **params):
        # type: (str, Any) -> List[MarketInstrument]
        payload = await self._request(
            method='GET',
            path=path,
            params=params,
        )
        return MarketInstrumentList.from_dict(payload).instruments


class MarketOrderBooksAPI(BaseTinkoffInvestmentsAPI):
    async def get(self, figi: FigiName, depth: int) -> OrderBook:
        payload = await self._request(
            method='GET',
            path='/market/orderbook',
            params={
                'figi': figi,
                'depth': depth,
            }
        )
        return OrderBook.from_dict(payload)


class MarketCandlesAPI(BaseTinkoffInvestmentsAPI):
    async def get(self, figi, dt_from, dt_to, interval):
        # type: (FigiName, datetime, datetime, CandleResolution) -> List[Candle]
        if not dt_from.tzinfo:
            dt_from = dt_from.replace(tzinfo=timezone.utc)
        if not dt_to.tzinfo:
            dt_to = dt_to.replace(tzinfo=timezone.utc)
        payload = await self._request(
            method='GET',
            path='/market/candles',
            params={
                'figi': figi,
                'from': dt_from.isoformat(),
                'to': dt_to.isoformat(),
                'interval': interval.value,
            },
        )
        return Candles.from_dict(payload).candles


class MarketAPI(BaseTinkoffInvestmentsAPI):
    def __init__(self, *args, **kwargs):
        super(MarketAPI, self).__init__(*args, **kwargs)
        self.instruments = MarketInstrumentsAPI(*args, **kwargs)
        self.orderbooks = MarketOrderBooksAPI(*args, **kwargs)
        self.candles = MarketCandlesAPI(*args, **kwargs)
